@@ -4,6 +4,7 @@ import os
 import sys
 import time
 import shutil
+import signal
 import atexit
 import socket
 import functools
@@ -338,6 +339,7 @@ class OsCommandPlugin(Plugin):
     '''
     param_type = dict
     inner_types = {
+                    'shell': {'required': False, 'type': bool},
                     'ignore_error': {'required': False, 'type': bool},
                     'init': {'required': False, 'type': int},
                     'background': {'required': False, 'type': bool},
@@ -353,8 +355,14 @@ class OsCommandPlugin(Plugin):
         poll = self.process.poll()
 
         if poll and poll != 0 and not ignore_error:
-            command = ' '.join(command)
-            raise OSError(f"Command '{command}' exited with a non zero status code.")
+
+            if type(command) is list:
+                command = ' '.join(command)
+
+            stdout, stderr = self.process.communicate()
+
+            raise OSError(f"Command '{command}' exited with a non zero status code." +
+                          f"\n\nstdout: {stdout}\n\nstderr: {stderr}")
 
     def run(self) -> None:
         '''
@@ -363,13 +371,18 @@ class OsCommandPlugin(Plugin):
         init = self.param.get('init', 0)
 
         command = self.param['cmd']
+        shell = self.param.get('shell', False)
         timeout = self.param.get('timeout', 0)
         background = self.param.get('background', False)
 
         for ctr in range(len(command)):
             command[ctr] = str(command[ctr])
 
-        self.process = subprocess.Popen(command, cwd=self.path.parent, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
+        if shell:
+            command = ' '.join(command)
+
+        self.process = subprocess.Popen(command, cwd=self.path.parent, stdout=subprocess.PIPE,
+                                        stderr=subprocess.STDOUT, shell=shell, preexec_fn=os.setsid)
 
         if timeout > 0:
             self.process.communicate(timeout=timeout)
@@ -386,8 +399,11 @@ class OsCommandPlugin(Plugin):
         '''
         Stop the process if still existing.
         '''
-        if hasattr(self, 'process') and self.process and self.process.poll() is None:
-            self.process.kill()
+        try:
+            os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)
+
+        except ProcessLookupError:
+            pass
 
 
 class MkdirPlugin(Plugin):
@@ -608,6 +624,7 @@ class CleanupCommandPlugin(Plugin):
     '''
     param_type = dict
     inner_types = {
+                    'shell': {'required': False, 'type': bool},
                     'ignore_error': {'required': False, 'type': bool},
                     'timeout': {'required': False, 'type': int},
                     'cmd': {'required': True, 'type': list}
@@ -621,20 +638,31 @@ class CleanupCommandPlugin(Plugin):
         poll = self.process.poll()
 
         if poll and poll != 0 and not ignore_error:
-            command = ' '.join(command)
-            raise OSError(f"Command '{command}' exited with a non zero status code.")
+
+            if type(command) is list:
+                command = ' '.join(command)
+
+            stdout, stderr = self.process.communicate()
+
+            raise OSError(f"Command '{command}' exited with a non zero status code." +
+                          f"\n\nstdout: {stdout}\n\nstderr: {stderr}")
 
     def stop(self) -> None:
         '''
         Run the specified command on stop.
         '''
         command = self.param['cmd']
+        shell = self.param.get('shell', False)
         timeout = self.param.get('timeout', 0)
 
         for ctr in range(len(command)):
             command[ctr] = str(command[ctr])
 
-        self.process = subprocess.Popen(command, cwd=self.path.parent, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
+        if shell:
+            command = ' '.join(command)
+
+        self.process = subprocess.Popen(command, cwd=self.path.parent, stdout=subprocess.PIPE,
+                                        stderr=subprocess.STDOUT, shell=shell)
 
         if timeout > 0:
             self.process.communicate(timeout=timeout)
